import vmath
import animation
import pygame
import glsupport
import random

#type IDs for entities:
ENTITYTYPE_PIPER = 0
ENTITYTYPE_RAT = 1
ENTITYTYPE_KID = 2
ENTITYTYPE_PARENT = 3
ENTITYTYPE_NOTE = 4


ANIMATED_ENTITY_SCALING_FACTOR = 0.1
TICKS_PER_ANIMATION_FRAME = 3

class EntityBase(object):
    pass

class AnimatedEntity(EntityBase):
    """all animated sprites (piper, children, parents, rats, notes)
    animated sprites can move and can be graphically animated"""
    def __init__(self, animation, sprite_orientation):
        
        self.gl_animation = glsupport.GL_Animation(animation,
                                       ANIMATED_ENTITY_SCALING_FACTOR)
        self.pos = vmath.Vector(0, 0)
        self.vel = vmath.Vector(0, 0)
        self.do_special_thing = False #play pipe or drown
        self._animation_index = None
        self.set_animation(0)
        self.sprite_orientation = sprite_orientation
        
    def get_is_moving(self):
        return (abs(self.vel) > vmath.EPSILON)
        
    def set_animation(self, animation_key):
        if (animation_key != self._animation_index):
            self._animation_index = animation_key
            self._next_frame_index = 0
            self._next_frame_fraction = 0
            self._heading_on_idle = random.choice((True, False))
        
    def render(self):
        self.gl_animation.SetAnimation(self._animation_index)
        self.gl_animation.SetFrame(self._next_frame_index)
        
        self.gl_animation.pos.x = self.pos.x
        self.gl_animation.pos.y = self.pos.y 
        self.gl_animation.mirrored = self._get_heading()
        self.gl_animation.render()
        
        self._next_frame()
        
    def _next_frame(self):
        self._next_frame_fraction += 1
        
        if self._next_frame_fraction < TICKS_PER_ANIMATION_FRAME:
            return
        
        frame_count = self.gl_animation.GetCurrentAnimationLength()
        
        self._next_frame_fraction = 0
        self._next_frame_index = (self._next_frame_index + 1) % frame_count
        
        
    def _get_heading(self):
        if (self.vel.x < 0):
            heading = animation.DIRECTION_LEFT
        elif (self.vel.x > 0):
            heading = animation.DIRECTION_RIGHT
        else:
            heading = self._heading_on_idle
        return heading ^ self.sprite_orientation #this is an XOR
        
    def get_bounding_rect(self):
        w, h = self.gl_animation.GetCurrentAnmationFrameSize()
        #note: position will be the lower middle of the sprite (where the feet are)
        x = self.pos.x - w//2
        y = self.pos.y       
        return pygame.Rect(x, y ,w , h)
    
    def collides_with(self, other):
        rect1 = self.get_bounding_rect()
        rect2 = other.get_bounding_rect()
        return rect1.colliderect(rect2)


class Obstacle(EntityBase):
    """houses, towers, walls, ..."""
    
    def __init__(self, gl_foreground_image, gl_underground_image):
        self.pos = vmath.Vector(0, 0)
        self.gl_foreground_image = gl_foreground_image
        self.gl_underground_image = gl_underground_image
        
    def scale(self, scaling_factor):
        if self.gl_foreground_image != None:
            self.gl_foreground_image.set_scaling(scaling_factor)
        if self.gl_underground_image != None:
            self.gl_underground_image.set_scaling(scaling_factor)
        
    def changeWidth(self,new_width):
        if self.gl_foreground_image != None:
            self.gl_foreground_image.changeWidth(new_width)
        if self.gl_underground_image != None:
            self.gl_underground_image.changeWidth(new_width)
    
    def render(self):        
        if self.gl_foreground_image != None:
            self.gl_foreground_image.pos.x = self.pos.x
            self.gl_foreground_image.pos.y = self.pos.y 
            self.gl_foreground_image.render()
        if self.gl_underground_image != None:
            self.gl_underground_image.pos.x = self.pos.x
            self.gl_underground_image.pos.y = self.pos.y
            self.gl_underground_image.render()            
        #surface.fill(color, self.rectangle)
        
        