"""
Levels of the game.
There is one class for each level (derived from the Level base class).
The levels also double as an entity manager. 
"""
import entities
import random
import vmath
import animation

from glsupport import *
import globalconstants

import font2

ENTITY_VELOCITY = 0.5
PIPER_VELOCITY = 3
NOTE_VELOCITY = 1.5
NOTE_TIME_TO_FADE = 140
NOTE_RESPAWN = 5

LEVEL_W = 800
LEVEL_H = 600


class Level(object):
    """base class for all levels"""
    
    def __init__(self):
        self.entities = [] 
        self.entities_drowning = []
        self.textures = animation.create_texture_pool()
        
        self.town_map = TownMap()
        
        
        self.piper = entities.AnimatedEntity(self.textures[animation.SPRITE_PIPER],  
                                             animation.SPRITE_FACING_RIGHT)
        
        self.piper.type = animation.SPRITE_PIPER
        
        self.piper.pos.x = 370
        self.piper.pos.y = 30
        self.notes = []
        random.seed()
        
        self.font_engine = font2.Font()
        self._kills_open = 20
        self._note_tick = 0
        
        self.bg = GL_Texture("../img/bg.png")
    
    def tick(self, input):
        self._render(input)
        self._move(input)
    
    def _render(self, input):
        
        BeginGLRendering(self.piper.pos.x,self.piper.pos.y)
        
        self.bg.render(0, 50, 0, 800, 550, False)
        
        entities_to_render = self._create_entities_to_render()
        for entity in entities_to_render:        
            entity.render()
            
        #debug output:
        self.font_engine.draw_string(self._kills_open, 10, 10)
        
                    
        EndGLRendering()
            
    def _move(self, input):
        self._move_drowning()
        self._move_entities()
        self._move_piper(input)
        self._move_notes()
        
    def _move_drowning(self):
        
        for entity in self.entities_drowning:
            if entity.drown_step == 0:
                entity.ttl = 1 # this step doesn't use ttl
                entity.pos += entity.vel
                if not self.town_map.is_on_bridge(entity):
                    entity.drown_step = 5
                    entity.ttl = 8
                    
            elif entity.drown_step == 5:
                entity.ttl -= 1
                entity.pos += entity.vel
                entity.vel = vmath.Vector(entity.vel.x, entity.vel.y - 0.1)
                if entity.ttl < 1:
                    entity.set_animation(animation.ANIMATION_SPECIAL_1)
                    entity.drown_step = 10
                    entity.ttl = 8
                    self._kills_open -= 1
            
            elif entity.drown_step == 10:
                entity.ttl -= 1
                if entity.ttl < 1:
                    if (entity.type != animation.SPRITE_RAT):
                        entity.drown_step = 20
                        entity.set_animation(animation.ANIMATION_SPECIAL_2)
                        entity.ttl = 32
                    else:
                        entity.drown_step = 99
                        
            elif entity.drown_step == 20:
                entity.ttl -= 1
                if entity.ttl < 1:
                    entity.drown_step = 99
        
        self.entities_drowning = [entity for entity in self.entities_drowning if entity.drown_step < 99]
                    
        
    def _move_entities(self):
        entities_still_in_game = []
        
        for entity in self.entities:
            self.town_map.move_entity_in_town(entity)
            
            # small chance to start heading in random direction:
            if (random.random() < self.distractability):
                vel_x = random.uniform(-ENTITY_VELOCITY, ENTITY_VELOCITY)
                vel_y = random.uniform(-ENTITY_VELOCITY, ENTITY_VELOCITY)
                entity.vel = vmath.Vector(vel_x, vel_y)
                
            # head to piper if hit by note:
            for note in self.notes:    
                if entity.collides_with(note):
                    entity.vel = (self.piper.pos - entity.pos).get_vector_of_length(ENTITY_VELOCITY)
                    break
                    
            # set animation:    
            if (abs(entity.vel) < 0.2):
                entity.vel = vmath.Vector(0,0)
                entity.set_animation(animation.ANIMATION_IDLE)
            else:
                entity.set_animation(animation.ANIMATION_WALK)
                
            # check if entity on bridge:
            if self.town_map.is_on_bridge(entity):
                self.entities_drowning.append(entity)
                dir_left = vmath.Vector(-2.5, -2.5)
                dir_right = vmath.Vector(2.5, -2.5)
                entity.vel = random.choice((dir_left, dir_right))
                entity.drown_step = 0
                if (entity.type == animation.SPRITE_RAT):
                    entity.set_animation(animation.ANIMATION_SPECIAL_2)
                else:
                    entity.set_animation(animation.ANIMATION_WALK)
                
            else:
                entities_still_in_game.append(entity)
                
        self.entities = entities_still_in_game
        
                    
    def _move_notes(self):
        for note in self.notes:
            self.town_map.move_entity_in_town(note) # move note
            if note.last_move_success:
                note.ttl -= 1 # fade note
            else:
                note.ttl = 0 # note collided => remove now
        self.notes = [note for note in self.notes if note.ttl > 0] #remove faded notes
        
        
    def _move_piper(self, input):
        piper_vec = self.piper.pos
        mouse_vec = vmath.Vector(*input.mouse_pos)
        direction_vec = mouse_vec - piper_vec
        
        self._note_tick = (self._note_tick + 1) % NOTE_RESPAWN
        
        if (input.lmb):
            # LMB pressed => move piper towards cursor
            direction_vec = mouse_vec - piper_vec
            vel_vec = direction_vec.get_vector_of_length(PIPER_VELOCITY)
            self.piper.vel = vel_vec
            self.town_map.move_entity_in_town(self.piper) # piper is only moved during LMB down
            
        if (input.rmb) and (self._note_tick == 0):
            # RMB pressed => generate note moving towards cursor
            spawn_vec = piper_vec + vmath.Vector(0, 20)
            direction_vec = mouse_vec - spawn_vec      
            self._spawn_note(spawn_vec, direction_vec)
        
        if input.lmb and input.rmb:
            anim_index = animation.ANIMATION_SPECIAL_2
        elif input.lmb:
            anim_index = animation.ANIMATION_WALK
        elif input.rmb:
            anim_index = animation.ANIMATION_SPECIAL_1
        else:
            anim_index = animation.ANIMATION_IDLE 
        
        self.piper.set_animation(anim_index)
    
    def _spawn_note(self, pos_vec, direction_vec):
        newnote = entities.AnimatedEntity(self.textures[animation.SPRITE_NOTE], 
                                          animation.SPRITE_FACING_RIGHT)
        vel_vec = direction_vec.get_vector_of_length(NOTE_VELOCITY)
        newnote.pos = pos_vec
        newnote.ttl = NOTE_TIME_TO_FADE
        newnote.vel = vel_vec
        newnote.type = animation.SPRITE_NOTE
        self.notes.append(newnote)
    
    def _create_entities_to_render(self):
        # create a list of all entities, sorted by their drawing order
        entities_to_render = self.entities[:]
        entities_to_render.extend(self.town_map.all_things_to_draw)
        entities_to_render.append(self.piper)
        entities_to_render.extend(self.notes)
        entities_to_render.extend(self.entities_drowning)
        
        for entity in entities_to_render:
            entity.z = self.town_map.get_virtual_z(entity)
        
        entities_to_render.sort(cmp=lambda a, b: cmp(a.z, b.z))
        return entities_to_render
    
    def is_done(self):
        return (self._kills_open < 1)

class LevelOne(Level):
    """first level - catching rats"""
    
    def __init__(self):
        Level.__init__(self)
        
        self.distractability = 0.004
        
        for i in range(200):
            texture = self.textures[animation.SPRITE_RAT]
            
            entity = entities.AnimatedEntity(texture, 
                                             animation.SPRITE_FACING_LEFT)
            
            entity.pos = self.town_map.get_random_valid_location()
            
            entity.type = animation.SPRITE_RAT
            
            self.entities.append(entity)


class LevelTwo(Level):
    """second level - catching children"""
    
    def __init__(self):
        Level.__init__(self)
        
        self.distractability = 0.03
        
        for i in range(130):
            variant = random.choice((animation.SPRITE_KID_A,
                                    animation.SPRITE_KID_B,
                                    animation.SPRITE_KID_C))
            
            texture = self.textures[variant]
            
            entity = entities.AnimatedEntity(texture,
                                             animation.SPRITE_FACING_RIGHT)
            
            entity.pos = self.town_map.get_random_valid_location()
            
            entity.type = variant
            
            self.entities.append(entity)
 
 
       

   
class TownMap(object):
    """represents the town map (houses and background)"""
    def __init__(self):
        self.cornerTower = GL_Image( GL_Texture("../img/town/cornertower.png"))
        self.gateTower = GL_Image( GL_Texture("../img/town/gateTower.png"))
        self.housefront = GL_Image( GL_Texture("../img/town/housefront.png"))
        
        gl_texture_houseside = GL_Texture("../img/town/houseside.png")
        self.houseside_left = GL_Image(gl_texture_houseside)
        self.houseside_right = GL_Image(gl_texture_houseside)
        self.houseside_right.mirrored = True 
        self.wall_horizontal = GL_Image( GL_Texture("../img/town/wallsegment.png"))
        self.wall_vertical = GL_Image(GL_Texture("../img/town/verticalwall.png"))
        self.bridge = GL_Image(GL_Texture("../img/town/bridge.png"))
        
        #    x = obstacle without image
        #    b = bridge
        #    v = house from front
        #    < = house from left
        #    > = house from right
        #    t = tower without bridge
        #    g = gate
        #    w = water
        #    - = horizontal wall
        #    | = vertical wall
        
        self.villagemap = [
                          "t--------------t",
                          "|     v       <|",
                          "|> v     vvvv  |",
                          "|              |",
                          "| v <>     vv  |",
                          "|              |",
                          "|  <>    v    <|",
                          "|    v   v  < v|",
                          "|v          v  |",
                          "|              |",
                          "t------g-------t",
                          "wwwwwwwbwwwwwwww"
                          ]
        
        self.villagemap.reverse() # reverse map lines (in place), because GL coordinate system goes bottom up
        
        self.valid_spawn_tiles = []
        
        #init obstacle object according to village map:
        self.all_things_to_draw = []
                
        self.line_count = len(self.villagemap)
        self.col_count = len(self.villagemap[0])
        self.obstacle_w = RESOLUTION[0] / self.col_count
        self.obstacle_h = RESOLUTION[1] / self.line_count
           
        for line_number, line in zip(range(self.line_count), self.villagemap):
            for col_number, col in zip(range(self.col_count), line):
                if (col == " "):
                    self.valid_spawn_tiles.append((col_number, line_number))
                    continue
                #if (col == "x"): continue
                elif (col == "b"): obstacle = entities.Obstacle(self.bridge,None)
                elif (col == "v"): obstacle = entities.Obstacle(self.housefront,None)
                elif (col == "<"): obstacle = entities.Obstacle(self.houseside_left,None)
                elif (col == ">"): obstacle = entities.Obstacle(self.houseside_right,None)
                elif (col == "t"): obstacle = entities.Obstacle(self.cornerTower,None)
                elif (col == "g"): obstacle = entities.Obstacle(self.gateTower,None)
                #elif (col == "w"): obstacle = entities.Obstacle(GL_Image( obstacle_texture),None)
                elif (col == "-"): obstacle = entities.Obstacle(self.wall_horizontal,None)
                elif (col == "|"): obstacle = entities.Obstacle(self.wall_vertical,None)
                else: continue
                        
                obstacle_x = col_number * self.obstacle_w
                obstacle_y = line_number * self.obstacle_h
                
                
                obstacle.pos.x = obstacle_x + (self.obstacle_w / 2.0)
                obstacle.pos.y = obstacle_y
                
                #((col == "b") or (col == "w")) and - 1 or (obstacle_y + self.obstacle_h)
                
                #manual tweaking of location and size:
                
                if (col == "t") or (col == "g") or (col == "b"):
                    obstacle.scale(0.24)
                else:
                    obstacle.changeWidth( self.obstacle_w )
                
                if (col == "t") or (col == "g"):
                    obstacle.pos.y -= 20   
                
                if (col == "g"):
                    obstacle.pos.x -= 1.9
                
                if (col == "b"):                     
                    obstacle.pos.y -= 50
                    
                
                self.all_things_to_draw.append(obstacle)
                obstacle.type = col
        
        
        #TODO: pre-sort list to improve drawing performance
        
    def get_virtual_z(self, entity):
        """gets the z index for the entity"""
        real_y = entity.pos.y
        
        if ((real_y <= 2*self.obstacle_h) and
            (entity.type != "b") and (entity.type != "g") and 
            (entity.type != "-")):
            return (4 * self.obstacle_h) - real_y
        else:
            return - real_y
        
    def get_random_valid_location(self):
        tile = random.choice(self.valid_spawn_tiles)
        x_off = random.uniform(0.0, 1.0)
        y_off = random.uniform(0.0, 1.0)
        x = RESOLUTION[0] * (tile[0] + x_off) / self.col_count
        y = RESOLUTION[1] * (tile[1] + y_off) / self.line_count
        vec = vmath.Vector(x, y)
        return vec
        
        
    def move_entity_in_town(self, entity):
        new_pos_vec = entity.pos + entity.vel
        is_current_pos_valid = self.is_position_valid(entity.pos)
        is_new_pos_valid = self.is_position_valid(new_pos_vec)
        
        if (is_new_pos_valid) or (not is_current_pos_valid):
            entity.pos = new_pos_vec
            entity.last_move_success = True
        else:
            entity.last_move_success = False
        
        
    def is_position_valid(self, pos_vec):
        floortile = self._position_to_tile(pos_vec)
        if (floortile == " ") or (floortile == "b") or (floortile == "g"):
            return True
        else:
            return False
    
    def is_on_bridge(self, entity):
        floortile = self._position_to_tile(entity.pos)
        return (floortile == "b")
    
    def _position_to_tile(self, pos_vec):
        floortile_col = int(pos_vec.x / self.obstacle_w)
        floortile_line = int(pos_vec.y / self.obstacle_h)
        
        if (floortile_col < 0) or (floortile_col >= self.col_count) or (floortile_line < 0) or (floortile_line >= self.line_count):
            # entity would be moved outside of the town
            return "?"
        
        floortile = self.villagemap[floortile_line][floortile_col]
        return floortile
    